//---------------------------------------------------------
// RG XML DB Tools Collection
// Copyright © 2003 - 2009 Ronald D. Grindle
//---------------------------------------------------------

package com.qanavigation.XMLDB;

import com.qanavigation.Basics.Rslt;
import com.qanavigation.Basics.Rslt2DArray;
import com.qanavigation.Basics.RsltStatVal;
import com.qanavigation.Basics.RsltString;
import com.qanavigation.Basics.RsltTable;
import com.qanavigation.Basics.SFormatter;
import com.qanavigation.Basics.SetList;
import com.qanavigation.DBacc.DBacc2;
import com.qanavigation.XMLBasics.RsltXMLDoc;
import com.qanavigation.XMLBasics.RsltXMLNode;

/**
 * Service Class for XMLs in a DB.
 * XMLDB isA XMLDBbasics.
 * Service calls do not spoil the status of this object on failure. The actual status 
 * is found with the object returned.
 */
public class XMLDB extends XMLDBbasics 
{

	/**
	 * 
	 * @param p_DBAcc
	 * @param p_UIDservHostName
	 * @param p_port
	 */
	public XMLDB( DBacc2 p_DBAcc, String p_UIDservHostName, int p_port ) {

		super( p_DBAcc, p_UIDservHostName, p_port );
	}	

	/**
	 * Appends a (new) sub tree to an existing node in the DB
	 * @param p_UID
	 * @param p_Node
	 * @param p_wChk The existence of the parent node is checked
	 * @return The new UID of the top node of the tree
	 */
	public RsltString appSubTree( String p_UID, RsltXMLNode p_Node, boolean p_wChk ) {
		
		RsltString rStr = new RsltString( true );
		
		Rslt res;
		
		if( p_wChk ) {
			
			res = chkDoc( p_UID );
			
			rStr.checkRslt( res );
		}
		
		if( rStr.isOK() ) {
		
			startTrans();
			
			rStr = writeNodes( p_Node.m_xNode, p_UID );
			
			if( rStr.isOK() ) {
				
				commit();
			}
			else {
				
				rollBack();
			}
		}
		
		return rStr;
	}
	
	/**
	 * 
	 * @param p_pUID Parent UID.
	 * @param p_nodeName
	 * @param p_single Only one node expected?
	 * @return
	 */
	public Rslt deleteSubTree( String p_pUID, String p_nodeName, boolean p_single ) {
		
		Rslt res = new Rslt( true );
		
		int i, x;
		
		SetList uidLst;

		startTrans();

		uidLst = getUIDLstByParent( p_nodeName, p_pUID );

		if( res.checkRslt( uidLst, true ) ) {

			x = uidLst.size();

			if( p_single && ( x > 1 ) ) {

				res.setRsltStat( RsltStatVal.isAmbig );
			}
			else {
				
				for( i = 0; i < x; ++i ) {

					res.checkRslt( deleteNode( uidLst.get( i ) ) );
				}
			}

			if( res.isOK() ) {

				res = commit();
			}
			else {

				rollBack();

				res.addRsltAddInfo( "'XMLDB.deleteSubTree( " + p_pUID + ", " + p_nodeName + " )' failed." );
			}		
		}

		return res;	
	}

	/**
	 * Exports an XML. The file name (w/o path) is retrieved from the DB.
	 * Fails if file name is not available.
	 * @param p_UID
	 * @param p_ApplInfo
	 * @return
	 */
	public RsltString export2File( String p_UID, String p_ApplInfo ) {

		Rslt res;

		RsltTable fileInfo;
		RsltString fileName = new RsltString();

		fileInfo = getFileInfo( p_UID );

		if( fileInfo.notOKorEmpty() ) {

			fileName.setRsltStat( fileInfo );
			fileName.addRsltAddInfo( "'rg.XMLDB.XMLDB.export2File([1])' failed." );
		}
		else {

			fileName.setString( fileInfo.getCell( 0, 0 ) );
			
			res = export2File( p_UID, fileName.getString(), p_ApplInfo );

			fileName.setRsltStat( res );
		}

		return fileName;
	}

	/**
	 * Exports an XML to a file.
	 * @param p_UID
	 * @param p_FileName
	 * @param p_ApplInfo is used for comment  "Generated by " + p_ApplInfo
	 * @return
	 */
	public Rslt export2File( String p_UID, String p_FileName, String p_ApplInfo ) {

		Rslt res = new Rslt();
		XMLDBNode xNode;

		xNode = getXMLsubTree( new RsltString( p_UID ) );

		if( xNode.isOK() ) {
 
			//<!-- Generated by QA Navigator Test Case Editor V1.0.0.0-->
			String genCom = "Generated by " + p_ApplInfo;
			xNode.addComment( genCom );
		}

		if( xNode.isOK() ) {

			xNode.m_RxDoc.save( p_FileName );
		}

		res.setRsltStat( xNode );

		if( xNode.notOK() ) {

			res.addRsltAddInfo( "'rg.XMLDB.XMLDB.export2File([2])' failed." );
		}

		return res;
	}

	/**
	 * Gets the list of attributes for a node as a string.
	 * @param p_UID
	 * @return
	 */
	public RsltString getAttrString( String p_UID ) { 

		int i, x;

		RsltTable attrTbl;

		RsltString attrString = new RsltString();
		String val = "";

		attrTbl = getNodeAttrs( p_UID );

		if( attrString.checkRslt( attrTbl ) ) {

			x = attrTbl.getRowNum();
		
			for( i = 0; i < x; i++ ) {

				val += " " + attrTbl.getCell( 0, i );
				val += "=\"" + attrTbl.getCell( 1, i ) + "\"";
			}

			attrString.setString( val );
		}

		if( attrString.notOK() ) {

			attrString.addRsltAddInfo( "'rg.XMLDB.XMLDB.getAttrString(..)' failed." );
		}

		return attrString;
	}

	/**
	 * Gets all documents of the given type where the content or a title has the requested value.
	 * @param p_ElemType
	 * @param p_Text
	 * @param p_CaseSens
	 * @return
	 */
	public SetList getDocUIDsByVal( String p_ElemType, String p_Text, boolean p_CaseSens ) {
		
		int i, x;
		
		SetList UIDLst;
		SetList aUIDLst;
		SetList pUIDLst = new SetList();
		
		pUIDLst.setRsltStat( RsltStatVal.isEmpty );
		
		String whereStmnt;
		
		String txt;
		RsltString rUID;
		RsltString rName;
		String name;
		String cUID = "";
		String UID = "";
		String tUID = "";
		
		// MySQL does not give a fuck if the field is CHAR
		if( p_CaseSens ) {
			
			txt = p_Text;
		}
		else {
			
			txt = p_Text.toLowerCase();
		}
		
		txt = m_DB.escapeStr2( SFormatter.encodeHTML( txt ) );
		
		txt = "'%" + txt + "%'";
		
		//select UIDpNode from NodeVal where Value LIKE '%a%';	
		
		if( p_CaseSens ) {
			
			whereStmnt = XMLDBDefs.c_ColNmValue;
		}
		else {
			
			whereStmnt = "LOWER( " + XMLDBDefs.c_ColNmValue + " ) ";
		}			
		
		whereStmnt += " LIKE " + txt;
		
		if( !( p_CaseSens ) ) {		
		
			whereStmnt += " COLLATE latin1_general_ci";
		}
		
		UIDLst = m_DB.findAllByVal( XMLDBDefs.c_TblNmNodeVal, XMLDBDefs.c_ColNmParentUID, whereStmnt );
		
		whereStmnt += " AND " + XMLDBDefs.c_ColNmName + " = 'Title'";
		
		aUIDLst = m_DB.findAllByVal( XMLDBDefs.c_TblNmAttrib, XMLDBDefs.c_ColNmParentUID, whereStmnt );
		
		UIDLst.checkRslt( aUIDLst, true );
		
		UIDLst.merge( aUIDLst );
		
		x = UIDLst.size();
		
		for( i = 0; i < x; ++i ) {
			
			cUID = UIDLst.getString( i );
		
			rUID = getTopNodeUID( cUID );
			
			if( UIDLst.checkRslt( rUID ) ) {
				
				tUID = rUID.getString();
			
				if( UID.compareTo( tUID ) != 0 ) {
				
					UID = tUID;
					
					rName = getNodeName( UID );
					
					name = rName.getString();
					
					if( p_ElemType.compareTo( name ) == 0 ) {
					
						pUIDLst.add( UID );
					}
				}
			}
		}
		
		pUIDLst.distinct();
		
//UIDLst.addRsltAddInfo( "EscTxt #" + txt + "#" );		
		
		return pUIDLst;
	}

	/**
	 * Gets the list of attributes for a node as table.
	 * @param p_UID
	 * @return RsltTable; Cols: Name, Value.
	 */
	public RsltTable getNodeAttrs( RsltString p_UID ) { 

		RsltTable attrTbl;

		String uIDstr = p_UID.getString();

		attrTbl = getNodeAttrs( uIDstr );

		if( attrTbl.notOK() ) {

			attrTbl.addRsltAddInfo( "'rg.XMLDB.XMLDB.getAttrString(..)' failed." );
		}

		return attrTbl;
	}

	/**
	 * Gets the list of nodes, which match the given attribute values.
	 * Requires the ElemType (name of the top node, Empty string for any)
	 * plus a Rslt2DArray with attribute names and values. 
	 * @param p_ElemType
	 * @param p_AttrSet
 	 * @return RsltTable; Col sequence: UID, Name, Attribs.
	 */
	public RsltTable getNodeLstByAttrs( String p_ElemType, Rslt2DArray p_AttrSet ) {
		
		SetList UIDLst;
		RsltTable XMLTbl;

		UIDLst = getNodesByAttr( p_ElemType, p_AttrSet );

		if( UIDLst.notOK() ) {

			XMLTbl = new RsltTable();

			XMLTbl.setRsltStat( UIDLst );
		}
		else {

			XMLTbl = prepNodeLst( p_ElemType, UIDLst );
		}

		if( XMLTbl.notOK() ) {

			XMLTbl.addRsltAddInfo( "'rg.XMLDB.XMLDB.getNodeLstByAttrs(..)' failed." );
		}

		return XMLTbl;
	}

	/**
	 * Gets the value of a node.
	 * @param p_UID
	 * @return
	 */
	public RsltString getNodeVal( RsltString p_UID ) { 

		RsltString val;

		String uIDstr = p_UID.getString();

		val = getNodeVal( uIDstr );
		
		if( val.notOK() ) {

			val.addRsltAddInfo( "'rg.XMLDB.XMLDB.getNodeVal(..)' failed." );
		}

		return val;
	}

	/**
	 * 
	 * @param p_ElemType
	 * @param p_ParUID
	 * @return
	 */
	public RsltString getNodeValByParent( String p_ElemType, String p_ParUID ) {

		RsltString val;

		SetList lst;

		lst = getUIDLstByParent( p_ElemType, p_ParUID );

		if( lst.size() != 1 ) {

			val = new RsltString();

			val.setInError( "'" + p_ElemType + "' for '" + p_ParUID + "' was not found or is not unique." );
		}
		else {

			String UID = lst.getString(0);

			val = getNodeVal( UID );
		}
					
		if( val.notOK() ) {

			val.addRsltAddInfo( "'rg.XMLDB.XMLDB.getNodeValByParent(..)' failed." );
		}

		return val;
	}

	/**
	 * 
	 * @param p_parentUID
	 * @param p_SubNodeName
	 * @param p_attrLst
	 * @return
	 */
	public RsltTable getSubNodesByAttr( String p_parentUID, String p_SubNodeName, Rslt2DArray p_attrLst ) {
		
		RsltTable uidLst;
		
		String[] row = new String[2];
		
		row[0] = XMLDBDefs.c_ColNmParentUID;
		row[1] = p_parentUID;
	
		p_attrLst.add( row );
		
		uidLst = getNodeLstByAttrs( p_SubNodeName, p_attrLst );
		
		return uidLst;
	}
	
	/**
	 * @return RsltTable; Col sequence: UID, Node Name, Attrib String.
	 */
	public RsltTable getTopNodeLst( ) {
		
		RsltTable XMLTbl;

		String empty = "";

		XMLTbl = getTopNodeLst( empty );

		if( XMLTbl.notOK() ) {

			XMLTbl.addRsltAddInfo( "'rg.XMLDB.XMLDB.getTopNodeLst([2])' failed." );
		}

		return XMLTbl;
	}
	
	/**
	 * Gets the list of top XML nodes.
	 * Requires the ElemType (name of the top node element).
	 * Returns a table with UID, node name and the assigned attributes.
	 * @param p_ElemType
	 * @return RsltTable; Col sequence: UID, Node Name, Attrib String.
	 */
	public RsltTable getTopNodeLst( String p_ElemType ) {
		
		SetList UIDLst;
		RsltTable XMLTbl;

		UIDLst = getUIDLstByParent( p_ElemType, XMLDBDefs.c_DummyTopUID );

		if( UIDLst.notOK() ) {

			XMLTbl = new RsltTable();
			XMLTbl.setRsltStat( UIDLst );
		}
		else {

			XMLTbl = prepNodeLst( p_ElemType, UIDLst );
		}

		if( XMLTbl.notOK() ) {

			XMLTbl.addRsltAddInfo( "'rg.XMLDB.XMLDB.getTopNodeLst([1])' failed." );
		}

		return XMLTbl;
	}

	/**
	 * 
	 * @param p_TopElemType
	 * @param p_ElemType
	 * @param p_AttrSet
	 * @return
	 */
	public SetList getTopNodesByAttr( String p_TopElemType, String p_ElemType, Rslt2DArray p_AttrSet ) {

		int i, x;
		SetList nodes;
		SetList tNodes;
		RsltString UID;

		UID = new RsltString();

		UID.setRsltStat( RsltStatVal.isEmpty, "" );

		nodes = getNodesByAttr( p_ElemType, p_AttrSet );

		if( nodes.notOK() ) {

			tNodes = new SetList();
			tNodes.setRsltStat( nodes );
		}
		else {

			x = nodes.size();

			tNodes = new SetList();

			for( i = 0; i < x; ++i ) {

				UID = getTopNodeUID( p_TopElemType, nodes.getString(i) );

				// may not be found, thus notOKs are ignored
				if( UID.isOK() ) {

					tNodes.add( UID.getString() );
				}
			}

			tNodes.setRsltStat( UID );
		}

		if( tNodes.notOK() ) {

			tNodes.addRsltAddInfo( "'rg.XMLDB.XMLDB.getTopNodesByAttr(..)' failed." );
		}

		return tNodes;
	}

	/**
	 * Gets the UID of the top node of the document, to which the node with the UID belongs to.
	 * @param p_UID
	 * @return
	 */
	public RsltString getTopNodeUID( String p_UID ) {

		RsltString parUID;

		RsltString rUID = new RsltString( p_UID );

		parUID = getTopNodeUID( rUID );

		if( parUID.notOK() ) {

			parUID.addRsltAddInfo( "'rg.XMLDB.XMLDB.getTopNode([1])' failed." );
		}

		return parUID;
	}

	/**
	 * Gets the UID of the top node of the specified element type (if there is) of 
	 * the document, to which the node with the UID bleongs to. Walks up the tree 
	 * until it finds the requested element type.  
	 * @param p_TopElemType
	 * @param p_UID
	 * @return The UID.
	 */
	public RsltString getTopNodeUID( String p_TopElemType, String p_UID ) {

		RsltString parUID;

		RsltString UID = new RsltString( p_UID );
		RsltString tNodeELem = new RsltString( p_TopElemType );

		parUID = getTopNodeUID( tNodeELem, UID );

		if( parUID.notOK() ) {

			parUID.addRsltAddInfo( "'rg.XMLDB.XMLDB.getTopNode([2])' failed." );
		}

		return parUID;
	}

	/**
	 * Gets the subtree, starting from the node specified with the UID, which can be the
	 * entire XML document, if the UID points to it.
	 * @param p_UID
	 * @return
	 */
	public XMLDBNode getXMLsubTree( RsltString p_UID ) { 

		XMLDBNode xNode;

		String uIDstr = p_UID.getString();

		xNode = getXMLsubTree( uIDstr );

		if( xNode.notOK() ) {

			xNode.addRsltAddInfo( "'rg.XMLDB.XMLDB.getXMLsubTree([2])' failed." );
		}
		
		return xNode;
	}

	/**
	 * Gets the subtree, starting from the node specified with the UID, which can be the
	 * entire XML document, if the UID points to it.
	 * @param p_UID
	 * @return The requested.
	 */
	public XMLDBNode getXMLsubTree( String p_UID ) { 

		XMLDBNode xNode;

		xNode = new XMLDBNode( this, p_UID );
		
		if( xNode.isOK() ) {
			
			xNode.gatherSubTree( );
		}

		if( xNode.notOK() ) {

			xNode.addRsltAddInfo( "'rg.XMLDB.XMLDB.getXMLsubTree([1])' failed." );
		}
		
		return xNode;
	}
	
	/**
	 * Imports an XML from the given file to the DB. 
	 * @param p_XMLfile
	 * @return The UID of the XML
	 */
	public RsltString importXMLFile( String p_XMLfile ){

		RsltString UID = new RsltString();
		String UIDstr;
				
		RsltXMLDoc xDoc = new RsltXMLDoc( p_XMLfile );

		if( xDoc.notOK() ) {

			UID = new RsltString();
			UID.setRsltStat( xDoc );
		}
		else {

			UID = importXML( xDoc.m_xDoc );
		}

		if( UID.isOK() ) {

			UIDstr = UID.getString();

			writeFileInfo( UIDstr, p_XMLfile );

			if( notOK() ) {

				UID.setRsltStat( this );
			}
		}

		if( UID.notOK() ) {

			UID.addRsltAddInfo( "'rg.XMLDB.XMLDB.importXMLFile(..)' failed." );
		}

		return UID;
	}
	
	// -- Adds additional information to the node list.
	protected RsltTable prepNodeLst( String p_ElemType, SetList p_UIDLst ) {
	
		boolean getElemName = false;
	
		RsltTable XMLTbl;
		RsltString attribStr;
		String UIDstr;
		String[] rowVal;

		int i, x;

		if( p_UIDLst.notOK() ) {
		
			XMLTbl = new RsltTable();
			XMLTbl.checkRslt( p_UIDLst );
		}
		else {
			
			if( p_ElemType.length() < 1 ) {
				
				getElemName = true; 
			}

			//String[] xMLColNames = { "UID", "Node Name", "Attribs" /*, "Value"*/ };

			x = p_UIDLst.size();

			XMLTbl = new RsltTable( 3, x /*, xMLColNames*/ );
			
			for( i = 0; i < x; i++ ) {

				rowVal = new String[3];

				UIDstr = p_UIDLst.getString(i);

				attribStr = getAttrString( UIDstr );
				
				if( getElemName ) {
					
					p_ElemType = getNodeName( UIDstr ).getString();
				}

				rowVal[0] = p_UIDLst.getString(i);
				rowVal[1] = p_ElemType;
				rowVal[2] = attribStr.getString();

				XMLTbl.add( rowVal );

				/*
				XMLTbl.setCell( 0, i, (String)p_UIDLst[ i ] );

				//XMLTbl.setCell( 1, i, p_nodeTbl.getCell( 1, i ) );

				XMLTbl.setCell( 2, i, attribStr );
				*/
			}

			XMLTbl.setOK();
		}

		return XMLTbl;
	}
	
	/** 
	 * Updates the attribute, if it already existed, or sets it.
	 * @param p_NodeUID
	 * @param p_AttrName
	 * @param p_AttrVal
	 * @return The state of the update.
	 */
	public Rslt setAttrVal( String p_NodeUID, String p_AttrName, String p_AttrVal  ) { 

		Rslt res;
		
		startTrans();
		
		res = writeAttrNode2DB( p_NodeUID, p_AttrName, p_AttrVal );
		
/*		
res.addRsltAddInfo( "p_NodeUID: " + p_NodeUID );
res.addRsltAddInfo( "p_AttrName: " + p_AttrName );
res.addRsltAddInfo( "p_AttrVal: " + p_AttrVal );
*/
		
		if( res.isOK() ) {
			
			res = commit();
		}
		else {
			
			rollBack();
			
			res.addRsltAddInfo( "'XMLDB.setAttrVal( " + p_NodeUID + ", " + p_AttrName + ", " + p_AttrVal + ")' failed." );
		}		
		
		return res;
	}
	
	/**
	 * Updates or sets a sub node value.
	 * @param p_pUID
	 * @param p_nodeName
	 * @param p_newVal
	 * @return
	 */
	public Rslt setSubNodeVal( String p_pUID, String p_nodeName, String p_newVal ) {

		Rslt res = new Rslt( true );

		int x;

		RsltXMLNode nNode;

		SetList uidLst;

		startTrans();

		uidLst = getUIDLstByParent( p_nodeName, p_pUID );

		if( res.checkRslt( uidLst, true ) ) {

			x = uidLst.size();						

			if( x < 0 ) {				

				nNode = new RsltXMLNode( p_nodeName );

				nNode.setText( p_newVal );

				res.checkRslt( writeANode( nNode.m_xNode, p_pUID ) );	
			}
			else if( x > 1 ) {					

				res.setRsltStat( RsltStatVal.isAmbig );
			}
			else {		

				res = writeValNode2DB( uidLst.get( 0 ), p_newVal );
			}

			if( res.isOK() ) {

				res = commit();
			}
			else {

				rollBack();

				res.addRsltAddInfo( "'XMLDB.setSubNodeVal( " + p_pUID + ", " + p_nodeName + ", " + p_newVal + ")' failed." );
			}		
		}

		return res;		
	}
}
